@@ -1,13 +1,13 @@
 /* dd_rescue.c */
 /* 
- * dd_rescue copies your data from one file to another
- * files might as well be block devices, such as hd partitions
- * unlike dd, it does not necessarily abort on errors, but
+ * dd_rescue copies your data from one file to another.
+ * Files might as well be block devices, such as hd partitions.
+ * Unlike dd, it does not necessarily abort on errors but
  * continues to copy the disk, possibly leaving holes behind.
  * Also, it does NOT truncate the output file, so you can copy 
  * more and more pieces of your data, as time goes by.
- * So, this tool is suitable for rescueing data of crashed disk,
- * and that's the reason, it has been written by me.
+ * This tool is thus suitable for rescueing data of crashed disk,
+ * and that's the reason it has been written by me.
  *
  * (c) Kurt Garloff <garloff@suse.de>, 11/97, 10/99
  * Copyright: GNU GPL
@@ -26,7 +26,7 @@
 # define VERSION "(unknown)"
 #endif
 
-#define ID "$Id: dd_rescue.c,v 1.50 2005/02/14 00:39:44 garloff Exp $"
+#define ID "$Id: dd_rescue.c,v 1.59 2007/08/26 13:42:44 garloff Exp $"
 
 #ifndef SOFTBLOCKSIZE
 # define SOFTBLOCKSIZE 65536
@@ -55,12 +55,12 @@
 #include <sys/time.h>
 #include <sys/stat.h>
 
-int softbs, hardbs;
+int softbs, hardbs, syncfreq;
 int maxerr, nrerr, reverse, dotrunc, abwrerr, sparse, nosparse;
 int verbose, quiet, interact, force;
 void* buf;
-char *lname, *iname, *oname;
-off_t ipos, opos, xfer, lxfer, sxfer, fxfer, maxxfer;
+char *lname, *iname, *oname, *bbname = NULL;
+off_t ipos, opos, xfer, lxfer, sxfer, fxfer, maxxfer, init_opos;
 
 int ides, odes, identical, pres;
 int o_dir_in, o_dir_out;
@@ -71,9 +71,17 @@ struct timeval starttime, lasttime, currenttime;
 struct timezone tz;
 clock_t startclock;
 
-const char* up = "\x1b[A"; //] 
-const char* down = "\n";
-const char* right = "\x1b[C"; //]
+#ifndef UP
+# define UP "\x1b[A"
+# define DOWN "\n"
+# define RIGHT "\x1b[C"
+#endif
+
+const char* up = UP;
+const char* down = DOWN;
+const char* right = RIGHT;
+
+int lastsparse = 0;
 
 inline float difftimetv(const struct timeval* const t2, 
 			const struct timeval* const t1)
@@ -178,17 +186,21 @@ void doprint(FILE* const file, const int bs, const clock_t cl,
 			100.0*(cl-startclock)/(CLOCKS_PER_SEC*t1));
 }
 
-void printstatus(FILE* const file1, FILE* const file2, 
+void printstatus(FILE* const file1, FILE* const file2,
 		 const int bs, const int sync)
 {
 	float t1, t2; 
 	clock_t cl;
-	int err = 0;
+
+	if (file1 == stderr || file1 == stdout) 
+		fprintf(file1, "%s%s%s", up, up, up);
+	if (file2 == stderr || file2 == stdout) 
+		fprintf(file2, "%s%s%s", up, up, up);
 
 	if (sync) {
-		err = fsync(odes);
+		int err = fsync(odes);
 		if (err)
-			fplog(stderr, "dd_rescue: (warning): %s (%.1fk): %s!\n", 
+			fplog(stderr, "dd_rescue: (warning): %s (%.1fk): %s!    \n", 
 			      oname, (float)ipos/1024, strerror(errno));
 	}
 
@@ -197,11 +209,6 @@ void printstatus(FILE* const file1, FILE* const file2,
 	t2 = difftimetv(&currenttime, &lasttime);
 	cl = clock();
 
-	if (file1 == stderr || file1 == stdout) 
-		fprintf(file1, "%s%s%s", up, up, up);
-	if (file2 == stderr || file2 == stdout) 
-		fprintf(file2, "%s%s%s", up, up, up);
-
 	if (file1) 
 		doprint(file1, bs, cl, t1, t2, sync);
 	if (file2)
@@ -212,6 +219,17 @@ void printstatus(FILE* const file1, FILE* const file2,
 	}
 }
 
+void savebb(int block)
+{
+	FILE *bbfile;
+	fplog(stderr, "Bad block: %d\n", block);
+	if (bbname == NULL)
+		return;
+	bbfile = fopen(bbname, "a");
+	fprintf(bbfile, "%d\n", block);
+	fclose(bbfile);
+}
+
 void printreport()
 {
 	/* report */
@@ -222,11 +240,27 @@ void printreport()
 	printstatus(report, logfd, 0, 1);
 }
 
+int mayexpandfile()
+{	
+	struct stat st;
+	off_t maxopos = opos;
+	if (init_opos > opos)
+		maxopos = init_opos;
+	stat(oname, &st);
+	if (!S_ISREG(st.st_mode))
+		return 0;
+	if (st.st_size < maxopos)
+		return truncate(oname, maxopos);
+	else 
+		return 0;		
+}
+
 int cleanup()
 {
 	int rc, errs = 0;
 	if (odes != -1) {
-		/* Make sure, the output file is expanded to the last (first) position */
+		/* Make sure, the output file is expanded to the last (first) position
+	 	 * FIXME: 0 byte writes do NOT expand file */
 		pwrite(odes, buf, 0, opos);
 		rc = fsync(odes);
 		if (rc) {
@@ -240,6 +274,11 @@ int cleanup()
 			      oname, (float)opos/1024, strerror(errno));
 			++errs;
 		}
+		if (sparse)
+			rc = mayexpandfile();
+			if (rc)
+				fplog(stderr, "dd_rescue: (warning): %s (%1.fk): %s!\n",
+				      oname, (float)opos/1024, strerror(errno));
 	}
 	if (ides != -1) {
 		rc = close(ides);
@@ -262,8 +301,8 @@ int blockiszero(const char* blk, const int ln)
 	unsigned long* ptr = (unsigned long*)blk;
 	while ((ptr-(unsigned long*)blk) < ln/sizeof(unsigned long))
 		if (*(ptr++)) 
-			return 0;
-	return 1;
+			return (sizeof(unsigned long)*(ptr-(unsigned long*)blk));
+	return ln;
 }
 
 inline ssize_t mypread(int fd, void* bf, size_t sz, off_t off)
@@ -330,8 +369,9 @@ int copyfile(const off_t max, const int bs)
 		(float)ipos/1024, (float)xfer/1024, (float)max/1024, bs,
 		down, down, down);
 #endif
-	/* expand file to the right length */
-	if (!o_chr) 
+	/* expand file to AT LEAST the right length 
+	 * FIXME: 0 byte writes do NOT expand file */
+	if (!o_chr)
 		pwrite(odes, buf, 0, opos);
 	while ( (!max || (max-xfer > 0))
 		&& ((!reverse) || (ipos > 0 && opos > 0)) ) {
@@ -385,12 +425,13 @@ int copyfile(const off_t max, const int bs)
 				errno = 0;
 				if (nosparse) {
 					ssize_t wr = 0;
-					errs += ((wr = writeblock(rd)) < rd ? 1: 0);
+					errs += ((wr = writeblock(toread)) < toread ? 1: 0);
+					lastsparse = 0;
 					if (wr < 0 && (errno == ENOSPC 
 						   || (errno == EFBIG && !reverse))) 
 						return errs;
-					if (rd != wr) {
-						fplog(stderr, "dd_rescue: (warning): assumption rd(%i) == wr(%i) failed! \n", rd, wr);	
+					if (toread != wr) {
+						fplog(stderr, "dd_rescue: (warning): assumption toread(%i) == wr(%i) failed! \n", toread, wr);	
 						/*
 						fplog(stderr, "dd_rescue: (warning): %s (%.1fk): %s!\n", 
 						      oname, (float)opos/1024, strerror(errno));
@@ -398,6 +439,7 @@ int copyfile(const off_t max, const int bs)
 					 	*/
 					}
 				}
+				savebb(ipos/bs);
 				fxfer += toread; xfer += toread;
 				if (reverse) { 
 					ipos -= toread; opos -= toread; 
@@ -414,8 +456,12 @@ int copyfile(const off_t max, const int bs)
 				/* But first: write available data and advance (optimization) */
 				if (rd > 0) {
 					ssize_t wr = 0; errno = 0;
-					if (!sparse || !blockiszero(buf, bs))
+					if (!sparse || blockiszero(buf, bs) < rd) {
 						errs += ((wr = writeblock(rd)) < rd ? 1: 0);
+						lastsparse = 0;
+					} else
+						lastsparse = 1;
+
 					if (!reverse) {
 						ipos += rd; opos += rd; 
 						sxfer += wr; xfer += rd;
@@ -440,7 +486,7 @@ int copyfile(const off_t max, const int bs)
 				old_xfer = xfer;
 				errs += (err = copyfile(new_max, hardbs));
 				/* EOF */
-				if (!err && old_xfer == xfer) 
+				if (!err && old_xfer == xfer)
 					return errs;
 				/*
 				if (reverse && rd) {
@@ -459,7 +505,7 @@ int copyfile(const off_t max, const int bs)
 				}
 				errno = 0;
 				/* EOF ? */      
-				if (!err && xfer == old_xfer) 
+				if (!err && xfer == old_xfer)
 					return errs;
 				if (verbose) 
 					fprintf(stderr, "dd_rescue: (info): ipos %.1fk promote to large bs again! \n%s%s%s",
@@ -469,8 +515,11 @@ int copyfile(const off_t max, const int bs)
 	      		/* errno == 0: We can write to disk */
       			if (rd > 0) {
 				ssize_t wr = 0;
-				if (!sparse || !blockiszero(buf, bs))
+				if (!sparse || blockiszero(buf, bs) < rd) {
 					errs += ((wr = writeblock(rd)) < rd ? 1: 0);
+					lastsparse = 0;
+				} else
+					lastsparse = 1;
 				sxfer += wr; xfer += rd;
 				if (reverse) { 
 					ipos -= rd; opos -= rd; 
@@ -489,10 +538,11 @@ int copyfile(const off_t max, const int bs)
 				}
 			} /* rd > 0 */
 		} /* errno */
-		if (!quiet && !(xfer % (16*softbs)) && (xfer % (512*softbs))) 
+
+		if (syncfreq && !(xfer % (syncfreq*softbs)))
+			printstatus((quiet? 0: stderr), 0, bs, 1);
+		else if (!quiet && !(xfer % (16*softbs)))
 			printstatus(stderr, 0, bs, 0);
-		if (!quiet && !(xfer % (512*softbs))) 
-			printstatus(stderr, 0, bs, 1);
 	} /* remain */
 	return errs;
 }
@@ -555,29 +605,31 @@ void printversion()
 void printhelp()
 {
 	printversion();
-	fprintf(stderr, "dd_rescue copies data from one file (or block device) to another\n");
+	fprintf(stderr, "dd_rescue copies data from one file (or block device) to another.\n");
 	fprintf(stderr, "USAGE: dd_rescue [options] infile outfile\n");
 	fprintf(stderr, "Options: -s ipos    start position in  input file (default=0),\n");
-	fprintf(stderr, "         -S opos    start position in output file (def=ipos);\n");
+	fprintf(stderr, "         -S opos    start position in output file (def=ipos),\n");
 	fprintf(stderr, "         -b softbs  block size for copy operation (def=%i),\n", SOFTBLOCKSIZE );
-	fprintf(stderr, "         -B hardbs  fallback block size in case of errs (def=%i);\n", HARDBLOCKSIZE );
-	fprintf(stderr, "         -e maxerr  exit after maxerr errors (def=0=infinite);\n");
-	fprintf(stderr, "         -m maxxfer maximum amount of data to be transfered (def=0=inf);\n");
-	fprintf(stderr, "         -l logfdile name of a file to log errors and summary to (def=\"\");\n");
-	fprintf(stderr, "         -r         reverse direction copy (def=forward);\n");
-	fprintf(stderr, "         -t         truncate output file (def=no);\n");
+	fprintf(stderr, "         -B hardbs  fallback block size in case of errs (def=%i),\n", HARDBLOCKSIZE );
+	fprintf(stderr, "         -e maxerr  exit after maxerr errors (def=0=infinite),\n");
+	fprintf(stderr, "         -m maxxfer maximum amount of data to be transfered (def=0=inf),\n");
+	fprintf(stderr, "         -y syncfrq frequency of fsync calls on outfile (def=512*softbs),\n");
+	fprintf(stderr, "         -l logfile name of a file to log errors and summary to (def=\"\"),\n");
+	fprintf(stderr, "         -o bbfile  name of a file to log bad blocks numbers (def=\"\"),\n");
+	fprintf(stderr, "         -r         reverse direction copy (def=forward),\n");
+	fprintf(stderr, "         -t         truncate output file (def=no),\n");
 #ifdef O_DIRECT
-	fprintf(stderr, "         -d/D       use O_DIRECT for input/output (def=no);\n");
+	fprintf(stderr, "         -d/D       use O_DIRECT for input/output (def=no),\n");
 #endif
-	fprintf(stderr, "         -w         abort on Write errors (def=no);\n");
+	fprintf(stderr, "         -w         abort on Write errors (def=no),\n");
 	fprintf(stderr, "         -a         spArse file writing (def=no),\n");
-	fprintf(stderr, "         -A         Always write blocks, zeroed if err (def=no);\n");
-	fprintf(stderr, "         -i         interactive: ask before overwriting data (def=no);\n");
-	fprintf(stderr, "         -f         force: skip some sanity checks (def=no);\n");
-	fprintf(stderr, "         -p         preserve: preserve ownership / perms (def=no)\n");
+	fprintf(stderr, "         -A         Always write blocks, zeroed if err (def=no),\n");
+	fprintf(stderr, "         -i         interactive: ask before overwriting data (def=no),\n");
+	fprintf(stderr, "         -f         force: skip some sanity checks (def=no),\n");
+	fprintf(stderr, "         -p         preserve: preserve ownership / perms (def=no),\n");
 	fprintf(stderr, "         -q         quiet operation,\n");
-	fprintf(stderr, "         -v         verbose operation;\n");
-	fprintf(stderr, "         -V         display version and exit;\n");
+	fprintf(stderr, "         -v         verbose operation,\n");
+	fprintf(stderr, "         -V         display version and exit,\n");
 	fprintf(stderr, "         -h         display this help and exit.\n");
 	fprintf(stderr, "Note: Sizes may be given in units b(=512), k(=1024), M(=1024^2) or G(1024^3) bytes\n");
 	fprintf(stderr, "This program is useful to rescue data in case of I/O errors, because\n");
@@ -618,6 +670,7 @@ void breakhandler(int sig)
 int main(int argc, char* argv[])
 {
 	int c;
+	off_t syncsz = -1;
 	void **mp = (void **) &buf;
 
   	/* defaults */
@@ -632,7 +685,7 @@ int main(int argc, char* argv[])
 	ides = -1; odes = -1; logfd = 0; nrerr = 0; buf = 0;
 	i_chr = 0; o_chr = 0;
 
-	while ((c = getopt(argc, argv, ":rtfihqvVwaAdDpb:B:m:e:s:S:l:")) != -1) {
+	while ((c = getopt(argc, argv, ":rtfihqvVwaAdDpb:B:m:e:s:S:l:o:y:")) != -1) {
 		switch (c) {
 			case 'r': reverse = 1; break;
 			case 't': dotrunc = O_TRUNC; break;
@@ -654,9 +707,11 @@ int main(int argc, char* argv[])
 			case 'B': hardbs = (int)readint(optarg); break;
 			case 'm': maxxfer = readint(optarg); break;
 			case 'e': maxerr = (int)readint(optarg); break;
+			case 'y': syncsz = readint(optarg); break;
 			case 's': ipos = readint(optarg); break;
 			case 'S': opos = readint(optarg); break;
 			case 'l': lname = optarg; break;
+			case 'o': bbname = optarg; break;
 			case ':': fplog (stderr, "dd_rescue: (fatal): option %c requires an argument!\n", optopt); 
 				printhelp();
 				exit(11); break;
@@ -668,6 +723,7 @@ int main(int argc, char* argv[])
 		}
 	}
   
+	init_opos = opos;
 	if (optind < argc) 
 		iname = argv[optind++];
 	if (optind < argc) 
@@ -712,6 +768,14 @@ int main(int argc, char* argv[])
 		cleanup(); exit(15);
 	}
 
+	/* Set sync frequency */
+	if (syncsz == -1)
+		syncfreq = 512;
+	else if (syncsz == 0)
+		syncfreq = 0;
+	else
+		syncfreq = (syncsz + softbs - 1) / softbs;
+
 	/* Have those been set by cmdline params? */
 	if (ipos == (off_t)-1) 
 		ipos = 0;
@@ -741,7 +805,7 @@ int main(int argc, char* argv[])
 	if (ides < 0) {
 		fplog(stderr, "dd_rescue: (fatal): %s: %s\n", iname, strerror(errno));
 		cleanup(); exit(22);
-	};
+	}
 
 	/* Overwrite? */
 	/* Special case '-': stdout */
@@ -776,12 +840,19 @@ int main(int argc, char* argv[])
 		copyperm(ides, odes);
 			
 	check_seekable(ides, odes);
+
 	if (0 && i_chr && o_chr) {
 		fprintf(stderr, "dd_rescue: (fatal): Sorry, there is no support yet for non-seekable\n");
 		fprintf(stderr, "                    input and output. This will hopefully change soon ... \n");
 		exit(19);
 	}
 
+	if (o_chr) {
+		if (!nosparse)
+			fprintf(stderr, "dd_rescue: (warning): Don't use sparse writes for non-seekable output\n");
+		nosparse = 1; sparse = 0;
+	}
+
 	/* special case: reverse with ipos == 0 means ipos = end_of_file */
 	if (reverse && ipos == 0) {
 		ipos = lseek(ides, ipos, SEEK_END);
@@ -860,7 +931,6 @@ int main(int argc, char* argv[])
 	}
 
 	c = copyfile(maxxfer, softbs);
-
 	gettimeofday(&currenttime, NULL);
 	printreport();
 	cleanup();
